import pytest
from jsonschema_path import SchemaPath
from referencing.exceptions import Unresolvable

from openapi_spec_validator import OpenAPIV2SpecValidator
from openapi_spec_validator import OpenAPIV30SpecValidator
from openapi_spec_validator import OpenAPIV31SpecValidator
from openapi_spec_validator.validation.exceptions import OpenAPIValidationError


class TestLocalOpenAPIv2Validator:
    LOCAL_SOURCE_DIRECTORY = "data/v2.0/"

    def local_test_suite_file_path(self, test_file):
        return f"{self.LOCAL_SOURCE_DIRECTORY}{test_file}"

    @pytest.mark.parametrize(
        "spec_file",
        [
            "petstore.yaml",
        ],
    )
    def test_valid(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)
        schema_path = SchemaPath.from_dict(
            spec,
            base_uri=spec_url,
        )
        validator = OpenAPIV2SpecValidator(schema_path)

        validator.validate()

        assert validator.is_valid() == True

    @pytest.mark.parametrize(
        "spec_file",
        [
            "empty.yaml",
        ],
    )
    def test_validation_failed(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)
        validator = OpenAPIV2SpecValidator(spec, base_uri=spec_url)

        with pytest.raises(OpenAPIValidationError):
            validator.validate()

        assert validator.is_valid() == False

    @pytest.mark.parametrize(
        "spec_file",
        [
            "missing-reference.yaml",
        ],
    )
    def test_ref_failed(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)

        with pytest.raises(Unresolvable):
            OpenAPIV2SpecValidator(spec, base_uri=spec_url).validate()


class TestLocalOpenAPIv30Validator:
    LOCAL_SOURCE_DIRECTORY = "data/v3.0/"

    def local_test_suite_file_path(self, test_file):
        return f"{self.LOCAL_SOURCE_DIRECTORY}{test_file}"

    @pytest.mark.parametrize(
        "spec_file",
        [
            "petstore.yaml",
            "petstore-separate/spec/openapi.yaml",
            "parent-reference/openapi.yaml",
            "property-recursive.yaml",
            "read-only-write-only.yaml",
        ],
    )
    def test_valid(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)
        validator = OpenAPIV30SpecValidator(spec, base_uri=spec_url)

        validator.validate()

        assert validator.is_valid() == True

    @pytest.mark.parametrize(
        "spec_file",
        [
            "empty.yaml",
        ],
    )
    def test_failed(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)
        validator = OpenAPIV30SpecValidator(spec, base_uri=spec_url)

        with pytest.raises(OpenAPIValidationError):
            validator.validate()

        assert validator.is_valid() == False

    @pytest.mark.parametrize(
        "spec_file",
        [
            "property-missing-reference.yaml",
        ],
    )
    def test_ref_failed(self, factory, spec_file):
        spec_path = self.local_test_suite_file_path(spec_file)
        spec = factory.spec_from_file(spec_path)
        spec_url = factory.spec_file_url(spec_path)

        with pytest.raises(Unresolvable):
            OpenAPIV30SpecValidator(spec, base_uri=spec_url).validate()


@pytest.mark.network
class TestRemoteOpenAPIv30Validator:
    REMOTE_SOURCE_URL = (
        "https://raw.githubusercontent.com/OAI/OpenAPI-Specification/"
    )

    def remote_test_suite_file_path(self, test_file):
        return f"{self.REMOTE_SOURCE_URL}{test_file}"

    @pytest.mark.parametrize(
        "spec_file",
        [
            "f75f8486a1aae1a7ceef92fbc63692cb2556c0cd/examples/v3.0/"
            "petstore.yaml",
            "f75f8486a1aae1a7ceef92fbc63692cb2556c0cd/examples/v3.0/"
            "api-with-examples.yaml",
            "970566d5ca236a5ce1a02fb7d617fdbd07df88db/examples/v3.0/"
            "api-with-examples.yaml",
        ],
    )
    def test_valid(self, factory, spec_file):
        spec_url = self.remote_test_suite_file_path(spec_file)
        spec = factory.spec_from_url(spec_url)

        OpenAPIV30SpecValidator(spec, base_uri=spec_url).validate()


@pytest.mark.network
class TestRemoteOpenAPIv31Validator:
    REMOTE_SOURCE_URL = (
        "https://raw.githubusercontent.com/"
        "OAI/OpenAPI-Specification/"
        "d9ac75b00c8bf405c2c90cfa9f20370564371dec/"
    )

    def remote_test_suite_file_path(self, test_file):
        return f"{self.REMOTE_SOURCE_URL}{test_file}"

    @pytest.mark.parametrize(
        "spec_file",
        [
            "comp_pathitems.yaml",
            "info_summary.yaml",
            "license_identifier.yaml",
            "mega.yaml",
            "minimal_comp.yaml",
            "minimal_hooks.yaml",
            "minimal_paths.yaml",
            "path_no_response.yaml",
            "path_var_empty_pathitem.yaml",
            "schema.yaml",
            "servers.yaml",
            "valid_schema_types.yaml",
        ],
    )
    def test_valid(self, factory, spec_file):
        spec_url = self.remote_test_suite_file_path(
            f"tests/v3.1/pass/{spec_file}"
        )
        spec = factory.spec_from_url(spec_url)

        OpenAPIV31SpecValidator(spec, base_uri=spec_url).validate()

    @pytest.mark.parametrize(
        "spec_file",
        [
            "invalid_schema_types.yaml",
            "no_containers.yaml",
            "server_enum_empty.yaml",
            "servers.yaml",
            "unknown_container.yaml",
        ],
    )
    def test_failed(self, factory, spec_file):
        spec_url = self.remote_test_suite_file_path(
            f"tests/v3.1/fail/{spec_file}"
        )
        spec = factory.spec_from_url(spec_url)

        with pytest.raises(OpenAPIValidationError):
            OpenAPIV31SpecValidator(spec, base_uri=spec_url).validate()



from dataclasses import dataclass


@dataclass(frozen=True)
class SpecVersion:
    """
    Spec version designates the OAS feature set.
    """

    keyword: str
    major: str
    minor: str

    def __str__(self) -> str:
        return f"OpenAPIV{self.major}.{self.minor}"



"""OpenAIP spec validator schemas module."""
from functools import partial

from jsonschema.validators import Draft4Validator
from jsonschema.validators import Draft202012Validator
from lazy_object_proxy import Proxy

from openapi_spec_validator.schemas.utils import get_schema_content

__all__ = ["schema_v2", "schema_v3", "schema_v30", "schema_v31"]

get_schema_content_v2 = partial(get_schema_content, "2.0")
get_schema_content_v30 = partial(get_schema_content, "3.0")
get_schema_content_v31 = partial(get_schema_content, "3.1")

schema_v2 = Proxy(get_schema_content_v2)
schema_v30 = Proxy(get_schema_content_v30)
schema_v31 = Proxy(get_schema_content_v31)

# alias to the latest v3 version
schema_v3 = schema_v31

get_openapi_v2_schema_validator = partial(Draft4Validator, schema_v2)
get_openapi_v30_schema_validator = partial(Draft4Validator, schema_v30)
get_openapi_v31_schema_validator = partial(Draft202012Validator, schema_v31)

openapi_v2_schema_validator = Proxy(get_openapi_v2_schema_validator)
openapi_v30_schema_validator = Proxy(get_openapi_v30_schema_validator)
openapi_v31_schema_validator = Proxy(get_openapi_v31_schema_validator)



import string
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional
from typing import cast

from jsonschema._format import FormatChecker
from jsonschema.exceptions import ValidationError
from jsonschema.protocols import Validator
from jsonschema_path.paths import SchemaPath
from openapi_schema_validator import oas30_format_checker
from openapi_schema_validator import oas31_format_checker
from openapi_schema_validator.validators import OAS30Validator
from openapi_schema_validator.validators import OAS31Validator

from openapi_spec_validator.validation.exceptions import (
    DuplicateOperationIDError,
)
from openapi_spec_validator.validation.exceptions import ExtraParametersError
from openapi_spec_validator.validation.exceptions import (
    ParameterDuplicateError,
)
from openapi_spec_validator.validation.exceptions import (
    UnresolvableParameterError,
)

if TYPE_CHECKING:
    from openapi_spec_validator.validation.registries import (
        KeywordValidatorRegistry,
    )


class KeywordValidator:
    def __init__(self, registry: "KeywordValidatorRegistry"):
        self.registry = registry


class ValueValidator(KeywordValidator):
    value_validator_cls: Validator = NotImplemented
    value_validator_format_checker: FormatChecker = NotImplemented

    def __call__(
        self, schema: SchemaPath, value: Any
    ) -> Iterator[ValidationError]:
        with schema.resolve() as resolved:
            value_validator = self.value_validator_cls(
                resolved.contents,
                _resolver=resolved.resolver,
                format_checker=self.value_validator_format_checker,
            )
            yield from value_validator.iter_errors(value)


class OpenAPIV30ValueValidator(ValueValidator):
    value_validator_cls = OAS30Validator
    value_validator_format_checker = oas30_format_checker


class OpenAPIV31ValueValidator(ValueValidator):
    value_validator_cls = OAS31Validator
    value_validator_format_checker = oas31_format_checker


class SchemaValidator(KeywordValidator):
    def __init__(self, registry: "KeywordValidatorRegistry"):
        super().__init__(registry)

        self.schema_ids_registry: Optional[List[int]] = []

    @property
    def default_validator(self) -> ValueValidator:
        return cast(ValueValidator, self.registry["default"])

    def __call__(
        self, schema: SchemaPath, require_properties: bool = True
    ) -> Iterator[ValidationError]:
        if not hasattr(schema.content(), "__getitem__"):
            return

        assert self.schema_ids_registry is not None
        schema_id = id(schema.content())
        if schema_id in self.schema_ids_registry:
            return
        self.schema_ids_registry.append(schema_id)

        nested_properties = []
        if "allOf" in schema:
            all_of = schema / "allOf"
            for inner_schema in all_of:
                yield from self(
                    inner_schema,
                    require_properties=False,
                )
                if "properties" not in inner_schema:
                    continue
                inner_schema_props = inner_schema / "properties"
                inner_schema_props_keys = inner_schema_props.keys()
                nested_properties += list(inner_schema_props_keys)

        if "anyOf" in schema:
            any_of = schema / "anyOf"
            for inner_schema in any_of:
                yield from self(
                    inner_schema,
                    require_properties=False,
                )

        if "oneOf" in schema:
            one_of = schema / "oneOf"
            for inner_schema in one_of:
                yield from self(
                    inner_schema,
                    require_properties=False,
                )

        if "not" in schema:
            not_schema = schema / "not"
            yield from self(
                not_schema,
                require_properties=False,
            )

        if "items" in schema:
            array_schema = schema / "items"
            yield from self(
                array_schema,
                require_properties=False,
            )

        if "properties" in schema:
            props = schema / "properties"
            for _, prop_schema in props.items():
                yield from self(
                    prop_schema,
                    require_properties=False,
                )

        required = schema.getkey("required", [])
        properties = schema.get("properties", {}).keys()
        if "allOf" in schema:
            extra_properties = list(
                set(required) - set(properties) - set(nested_properties)
            )
        else:
            extra_properties = list(set(required) - set(properties))

        if extra_properties and require_properties:
            yield ExtraParametersError(
                f"Required list has not defined properties: {extra_properties}"
            )

        if "default" in schema:
            default = schema["default"]
            nullable = schema.get("nullable", False)
            if default is not None or nullable is not True:
                yield from self.default_validator(schema, default)


class SchemasValidator(KeywordValidator):
    @property
    def schema_validator(self) -> SchemaValidator:
        return cast(SchemaValidator, self.registry["schema"])

    def __call__(self, schemas: SchemaPath) -> Iterator[ValidationError]:
        for _, schema in schemas.items():
            yield from self.schema_validator(schema)


class ParameterValidator(KeywordValidator):
    @property
    def schema_validator(self) -> SchemaValidator:
        return cast(SchemaValidator, self.registry["schema"])

    def __call__(self, parameter: SchemaPath) -> Iterator[ValidationError]:
        if "schema" in parameter:
            schema = parameter / "schema"
            yield from self.schema_validator(schema)


class OpenAPIV2ParameterValidator(ParameterValidator):
    @property
    def default_validator(self) -> ValueValidator:
        return cast(ValueValidator, self.registry["default"])

    def __call__(self, parameter: SchemaPath) -> Iterator[ValidationError]:
        yield from super().__call__(parameter)

        if "default" in parameter:
            # only possible in swagger 2.0
            default = parameter.getkey("default")
            if default is not None:
                yield from self.default_validator(parameter, default)


class ParametersValidator(KeywordValidator):
    @property
    def parameter_validator(self) -> ParameterValidator:
        return cast(ParameterValidator, self.registry["parameter"])

    def __call__(self, parameters: SchemaPath) -> Iterator[ValidationError]:
        seen = set()
        for parameter in parameters:
            yield from self.parameter_validator(parameter)

            key = (parameter["name"], parameter["in"])
            if key in seen:
                yield ParameterDuplicateError(
                    f"Duplicate parameter `{parameter['name']}`"
                )
            seen.add(key)


class MediaTypeValidator(KeywordValidator):
    @property
    def schema_validator(self) -> SchemaValidator:
        return cast(SchemaValidator, self.registry["schema"])

    def __call__(
        self, mimetype: str, media_type: SchemaPath
    ) -> Iterator[ValidationError]:
        if "schema" in media_type:
            schema = media_type / "schema"
            yield from self.schema_validator(schema)


class ContentValidator(KeywordValidator):
    @property
    def media_type_validator(self) -> MediaTypeValidator:
        return cast(MediaTypeValidator, self.registry["mediaType"])

    def __call__(self, content: SchemaPath) -> Iterator[ValidationError]:
        for mimetype, media_type in content.items():
            yield from self.media_type_validator(mimetype, media_type)


class ResponseValidator(KeywordValidator):
    def __call__(
        self, response_code: str, response: SchemaPath
    ) -> Iterator[ValidationError]:
        raise NotImplementedError


class OpenAPIV2ResponseValidator(ResponseValidator):
    @property
    def schema_validator(self) -> SchemaValidator:
        return cast(SchemaValidator, self.registry["schema"])

    def __call__(
        self, response_code: str, response: SchemaPath
    ) -> Iterator[ValidationError]:
        # openapi 2
        if "schema" in response:
            schema = response / "schema"
            yield from self.schema_validator(schema)


class OpenAPIV3ResponseValidator(ResponseValidator):
    @property
    def content_validator(self) -> ContentValidator:
        return cast(ContentValidator, self.registry["content"])

    def __call__(
        self, response_code: str, response: SchemaPath
    ) -> Iterator[ValidationError]:
        # openapi 3
        if "content" in response:
            content = response / "content"
            yield from self.content_validator(content)


class ResponsesValidator(KeywordValidator):
    @property
    def response_validator(self) -> ResponseValidator:
        return cast(ResponseValidator, self.registry["response"])

    def __call__(self, responses: SchemaPath) -> Iterator[ValidationError]:
        for response_code, response in responses.items():
            yield from self.response_validator(response_code, response)


class OperationValidator(KeywordValidator):
    def __init__(self, registry: "KeywordValidatorRegistry"):
        super().__init__(registry)

        self.operation_ids_registry: Optional[List[str]] = []

    @property
    def responses_validator(self) -> ResponsesValidator:
        return cast(ResponsesValidator, self.registry["responses"])

    @property
    def parameters_validator(self) -> ParametersValidator:
        return cast(ParametersValidator, self.registry["parameters"])

    def __call__(
        self,
        url: str,
        name: str,
        operation: SchemaPath,
        path_parameters: Optional[SchemaPath],
    ) -> Iterator[ValidationError]:
        assert self.operation_ids_registry is not None

        operation_id = operation.getkey("operationId")
        if (
            operation_id is not None
            and operation_id in self.operation_ids_registry
        ):
            yield DuplicateOperationIDError(
                f"Operation ID '{operation_id}' for '{name}' in '{url}' is not unique"
            )
        self.operation_ids_registry.append(operation_id)

        if "responses" in operation:
            responses = operation / "responses"
            yield from self.responses_validator(responses)

        names = []

        parameters = None
        if "parameters" in operation:
            parameters = operation / "parameters"
            yield from self.parameters_validator(parameters)
            names += list(self._get_path_param_names(parameters))

        if path_parameters is not None:
            names += list(self._get_path_param_names(path_parameters))

        all_params = list(set(names))

        for path in self._get_path_params_from_url(url):
            if path not in all_params:
                yield UnresolvableParameterError(
                    "Path parameter '{}' for '{}' operation in '{}' "
                    "was not resolved".format(path, name, url)
                )
        return

    def _get_path_param_names(self, params: SchemaPath) -> Iterator[str]:
        for param in params:
            if param["in"] == "path":
                yield param["name"]

    def _get_path_params_from_url(self, url: str) -> Iterator[str]:
        formatter = string.Formatter()
        path_params = [item[1] for item in formatter.parse(url)]
        return filter(None, path_params)


class PathValidator(KeywordValidator):
    OPERATIONS = [
        "get",
        "put",
        "post",
        "delete",
        "options",
        "head",
        "patch",
        "trace",
    ]

    @property
    def parameters_validator(self) -> ParametersValidator:
        return cast(ParametersValidator, self.registry["parameters"])

    @property
    def operation_validator(self) -> OperationValidator:
        return cast(OperationValidator, self.registry["operation"])

    def __call__(
        self, url: str, path_item: SchemaPath
    ) -> Iterator[ValidationError]:
        parameters = None
        if "parameters" in path_item:
            parameters = path_item / "parameters"
            yield from self.parameters_validator(parameters)

        for field_name, operation in path_item.items():
            if field_name not in self.OPERATIONS:
                continue

            yield from self.operation_validator(
                url, field_name, operation, parameters
            )


class PathsValidator(KeywordValidator):
    @property
    def path_validator(self) -> PathValidator:
        return cast(PathValidator, self.registry["path"])

    def __call__(self, paths: SchemaPath) -> Iterator[ValidationError]:
        for url, path_item in paths.items():
            yield from self.path_validator(url, path_item)


class ComponentsValidator(KeywordValidator):
    @property
    def schemas_validator(self) -> SchemasValidator:
        return cast(SchemasValidator, self.registry["schemas"])

    def __call__(self, components: SchemaPath) -> Iterator[ValidationError]:
        schemas = components.get("schemas", {})
        yield from self.schemas_validator(schemas)


class RootValidator(KeywordValidator):
    @property
    def paths_validator(self) -> PathsValidator:
        return cast(PathsValidator, self.registry["paths"])

    @property
    def components_validator(self) -> ComponentsValidator:
        return cast(ComponentsValidator, self.registry["components"])

    def __call__(self, spec: SchemaPath) -> Iterator[ValidationError]:
        if "paths" in spec:
            paths = spec / "paths"
            yield from self.paths_validator(paths)
        if "components" in spec:
            components = spec / "components"
            yield from self.components_validator(components)



"""OpenAPI spec validator validation decorators module."""
import logging
from functools import wraps
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import TypeVar

from jsonschema.exceptions import ValidationError

from openapi_spec_validator.validation.caches import CachedIterable
from openapi_spec_validator.validation.exceptions import OpenAPIValidationError

Args = TypeVar("Args")
T = TypeVar("T")

log = logging.getLogger(__name__)


def wraps_errors(
    func: Callable[..., Any]
) -> Callable[..., Iterator[ValidationError]]:
    @wraps(func)
    def wrapper(*args: Any, **kwds: Any) -> Iterator[ValidationError]:
        errors = func(*args, **kwds)
        for err in errors:
            if not isinstance(err, OpenAPIValidationError):
                # wrap other exceptions with library specific version
                yield OpenAPIValidationError.create_from(err)
            else:
                yield err

    return wrapper


def wraps_cached_iter(
    func: Callable[[Args], Iterator[T]]
) -> Callable[[Args], CachedIterable[T]]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> CachedIterable[T]:
        result = func(*args, **kwargs)
        return CachedIterable(result)

    return wrapper


def unwraps_iter(
    func: Callable[[Args], Iterable[T]]
) -> Callable[[Args], Iterator[T]]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Iterator[T]:
        result = func(*args, **kwargs)
        return iter(result)

    return wrapper



from jsonschema_path.typing import Schema

from openapi_spec_validator.versions.consts import VERSIONS
from openapi_spec_validator.versions.datatypes import SpecVersion
from openapi_spec_validator.versions.finders import SpecVersionFinder


def get_spec_version(spec: Schema) -> SpecVersion:
    finder = SpecVersionFinder(VERSIONS)
    return finder.find(spec)



"""OpenAPI spec validator validation validators module."""
import logging
import warnings
from functools import lru_cache
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import cast

from jsonschema.exceptions import ValidationError
from jsonschema.protocols import Validator
from jsonschema_path.handlers import default_handlers
from jsonschema_path.paths import SchemaPath

from openapi_spec_validator.schemas import openapi_v2_schema_validator
from openapi_spec_validator.schemas import openapi_v30_schema_validator
from openapi_spec_validator.schemas import openapi_v31_schema_validator
from openapi_spec_validator.schemas.types import AnySchema
from openapi_spec_validator.validation import keywords
from openapi_spec_validator.validation.decorators import unwraps_iter
from openapi_spec_validator.validation.decorators import wraps_cached_iter
from openapi_spec_validator.validation.decorators import wraps_errors
from openapi_spec_validator.validation.registries import (
    KeywordValidatorRegistry,
)

log = logging.getLogger(__name__)


class SpecValidator:
    resolver_handlers = default_handlers
    keyword_validators: Mapping[str, Type[keywords.KeywordValidator]] = {
        "__root__": keywords.RootValidator,
    }
    root_keywords: List[str] = []
    schema_validator: Validator = NotImplemented

    def __init__(
        self,
        schema: AnySchema,
        base_uri: str = "",
        spec_url: Optional[str] = None,
    ) -> None:
        if spec_url is not None:
            warnings.warn(
                "spec_url parameter is deprecated. " "Use base_uri instead.",
                DeprecationWarning,
            )
            base_uri = spec_url
        self.base_uri = base_uri

        if isinstance(schema, SchemaPath):
            self.schema_path = schema
            self.schema = schema.contents()
        else:
            self.schema = schema
            self.schema_path = SchemaPath.from_dict(
                self.schema,
                base_uri=self.base_uri,
                handlers=self.resolver_handlers,
            )

        self.keyword_validators_registry = KeywordValidatorRegistry(
            self.keyword_validators
        )

    def validate(self) -> None:
        for err in self.iter_errors():
            raise err

    def is_valid(self) -> bool:
        error = next(self.iter_errors(), None)
        return error is None

    @property
    def root_validator(self) -> keywords.RootValidator:
        return cast(
            keywords.RootValidator,
            self.keyword_validators_registry["__root__"],
        )

    @unwraps_iter
    @lru_cache(maxsize=None)
    @wraps_cached_iter
    @wraps_errors
    def iter_errors(self) -> Iterator[ValidationError]:
        yield from self.schema_validator.iter_errors(self.schema)

        yield from self.root_validator(self.schema_path)


class OpenAPIV2SpecValidator(SpecValidator):
    schema_validator = openapi_v2_schema_validator
    keyword_validators = {
        "__root__": keywords.RootValidator,
        "components": keywords.ComponentsValidator,
        "default": keywords.OpenAPIV30ValueValidator,
        "operation": keywords.OperationValidator,
        "parameter": keywords.OpenAPIV2ParameterValidator,
        "parameters": keywords.ParametersValidator,
        "paths": keywords.PathsValidator,
        "path": keywords.PathValidator,
        "response": keywords.OpenAPIV2ResponseValidator,
        "responses": keywords.ResponsesValidator,
        "schema": keywords.SchemaValidator,
        "schemas": keywords.SchemasValidator,
    }
    root_keywords = ["paths", "components"]


class OpenAPIV30SpecValidator(SpecValidator):
    schema_validator = openapi_v30_schema_validator
    keyword_validators = {
        "__root__": keywords.RootValidator,
        "components": keywords.ComponentsValidator,
        "content": keywords.ContentValidator,
        "default": keywords.OpenAPIV30ValueValidator,
        "mediaType": keywords.MediaTypeValidator,
        "operation": keywords.OperationValidator,
        "parameter": keywords.ParameterValidator,
        "parameters": keywords.ParametersValidator,
        "paths": keywords.PathsValidator,
        "path": keywords.PathValidator,
        "response": keywords.OpenAPIV3ResponseValidator,
        "responses": keywords.ResponsesValidator,
        "schema": keywords.SchemaValidator,
        "schemas": keywords.SchemasValidator,
    }
    root_keywords = ["paths", "components"]


class OpenAPIV31SpecValidator(SpecValidator):
    schema_validator = openapi_v31_schema_validator
    keyword_validators = {
        "__root__": keywords.RootValidator,
        "components": keywords.ComponentsValidator,
        "content": keywords.ContentValidator,
        "default": keywords.OpenAPIV31ValueValidator,
        "mediaType": keywords.MediaTypeValidator,
        "operation": keywords.OperationValidator,
        "parameter": keywords.ParameterValidator,
        "parameters": keywords.ParametersValidator,
        "paths": keywords.PathsValidator,
        "path": keywords.PathValidator,
        "response": keywords.OpenAPIV3ResponseValidator,
        "responses": keywords.ResponsesValidator,
        "schema": keywords.SchemaValidator,
        "schemas": keywords.SchemasValidator,
    }
    root_keywords = ["paths", "components"]



from io import StringIO
from unittest import mock

import pytest

from openapi_spec_validator.__main__ import main


def test_schema_v2_detect(capsys):
    """Test schema v2 is detected"""
    testargs = ["./tests/integration/data/v2.0/petstore.yaml"]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v2.0/petstore.yaml: OK\n" in out


def test_schema_v31_detect(capsys):
    """Test schema v3.1 is detected"""
    testargs = ["./tests/integration/data/v3.1/petstore.yaml"]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v3.1/petstore.yaml: OK\n" in out


def test_schema_v31(capsys):
    """No errors when calling proper v3.1 file."""
    testargs = [
        "--schema",
        "3.1.0",
        "./tests/integration/data/v3.1/petstore.yaml",
    ]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v3.1/petstore.yaml: OK\n" in out


def test_schema_v30(capsys):
    """No errors when calling proper v3.0 file."""
    testargs = [
        "--schema",
        "3.0.0",
        "./tests/integration/data/v3.0/petstore.yaml",
    ]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v3.0/petstore.yaml: OK\n" in out


def test_schema_v2(capsys):
    """No errors when calling with proper v2 file."""
    testargs = [
        "--schema",
        "2.0",
        "./tests/integration/data/v2.0/petstore.yaml",
    ]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v2.0/petstore.yaml: OK\n" in out


def test_many(capsys):
    """No errors when calling with proper v2 and v3 files."""
    testargs = [
        "./tests/integration/data/v2.0/petstore.yaml",
        "./tests/integration/data/v3.0/petstore.yaml",
    ]
    main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v2.0/petstore.yaml: OK\n" in out
    assert "./tests/integration/data/v3.0/petstore.yaml: OK\n" in out


def test_errors_on_missing_description_best(capsys):
    """An error is obviously printed given an empty schema."""
    testargs = [
        "./tests/integration/data/v3.0/missing-description.yaml",
        "--schema=3.0.0",
    ]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert (
        "./tests/integration/data/v3.0/missing-description.yaml: Validation Error:"
        in out
    )
    assert "Failed validating" in out
    assert "'description' is a required property" in out
    assert "'$ref' is a required property" not in out
    assert "1 more subschemas errors" in out


def test_errors_on_missing_description_full(capsys):
    """An error is obviously printed given an empty schema."""
    testargs = [
        "./tests/integration/data/v3.0/missing-description.yaml",
        "--errors=all",
        "--schema=3.0.0",
    ]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert (
        "./tests/integration/data/v3.0/missing-description.yaml: Validation Error:"
        in out
    )
    assert "Failed validating" in out
    assert "'description' is a required property" in out
    assert "'$ref' is a required property" in out
    assert "1 more subschema error" not in out


def test_schema_unknown(capsys):
    """Errors on running with unknown schema."""
    testargs = [
        "--schema",
        "x.x",
        "./tests/integration/data/v2.0/petstore.yaml",
    ]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert "error: argument --schema" in err
    assert not out


def test_validation_error(capsys):
    """SystemExit on running with ValidationError."""
    testargs = [
        "--schema",
        "3.0.0",
        "./tests/integration/data/v2.0/petstore.yaml",
    ]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert (
        "./tests/integration/data/v2.0/petstore.yaml: Validation Error:" in out
    )
    assert "Failed validating" in out
    assert "'openapi' is a required property" in out


@mock.patch(
    "openapi_spec_validator.__main__.OpenAPIV30SpecValidator.validate",
    side_effect=Exception,
)
def test_unknown_error(m_validate, capsys):
    """SystemExit on running with unknown error."""
    testargs = [
        "--schema",
        "3.0.0",
        "./tests/integration/data/v2.0/petstore.yaml",
    ]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "./tests/integration/data/v2.0/petstore.yaml: Error:" in out


def test_nonexisting_file(capsys):
    """Calling with non-existing file should sys.exit."""
    testargs = ["i_dont_exist.yaml"]
    with pytest.raises(SystemExit):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "No such file: i_dont_exist.yaml\n" in out


def test_schema_stdin(capsys):
    """Test schema from STDIN"""
    spes_path = "./tests/integration/data/v3.0/petstore.yaml"
    with open(spes_path) as spec_file:
        spec_lines = spec_file.readlines()
    spec_io = StringIO("".join(spec_lines))

    testargs = ["--schema", "3.0.0", "-"]
    with mock.patch("openapi_spec_validator.__main__.sys.stdin", spec_io):
        main(testargs)
    out, err = capsys.readouterr()
    assert not err
    assert "stdin: OK\n" in out



from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import TypeVar

T = TypeVar("T")


class CachedIterable(Iterable[T], Generic[T]):
    """
    A cache-implementing wrapper for an iterator.
    Note that this is class is `Iterable[T]` rather than `Iterator[T]`.
    It should not be iterated by his own.
    """

    cache: List[T]
    iter: Iterator[T]
    completed: bool

    def __init__(self, it: Iterator[T]):
        self.iter = iter(it)
        self.cache = list()
        self.completed = False

    def __iter__(self) -> Iterator[T]:
        return CachedIterator(self)

    def __next__(self) -> T:
        try:
            item = next(self.iter)
        except StopIteration:
            self.completed = True
            raise
        else:
            self.cache.append(item)
            return item

    def __del__(self) -> None:
        del self.cache


class CachedIterator(Iterator[T], Generic[T]):
    """
    A cache-using wrapper for an iterator.
    This class is only constructed by `CachedIterable` and cannot be used without it.
    """

    parent: CachedIterable[T]
    position: int

    def __init__(self, parent: CachedIterable[T]):
        self.parent = parent
        self.position = 0

    def __next__(self) -> T:
        if self.position < len(self.parent.cache):
            item = self.parent.cache[self.position]
        elif self.parent.completed:
            raise StopIteration
        else:
            item = next(self.parent)

        self.position += 1
        return item



from typing import Any
from typing import Hashable
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Protocol
from typing import runtime_checkable

from openapi_spec_validator.validation.exceptions import OpenAPIValidationError


@runtime_checkable
class SupportsValidation(Protocol):
    def is_valid(self, instance: Mapping[Hashable, Any]) -> bool:
        ...

    def iter_errors(
        self,
        instance: Mapping[Hashable, Any],
        base_uri: str = "",
        spec_url: Optional[str] = None,
    ) -> Iterator[OpenAPIValidationError]:
        ...

    def validate(
        self,
        instance: Mapping[Hashable, Any],
        base_uri: str = "",
        spec_url: Optional[str] = None,
    ) -> None:
        ...



from jsonschema.exceptions import ValidationError

from openapi_spec_validator.exceptions import OpenAPISpecValidatorError


class ValidatorDetectError(OpenAPISpecValidatorError):
    pass


class OpenAPIValidationError(ValidationError):  # type: ignore
    pass


class ExtraParametersError(OpenAPIValidationError):
    pass


class ParameterDuplicateError(OpenAPIValidationError):
    pass


class UnresolvableParameterError(OpenAPIValidationError):
    pass


class DuplicateOperationIDError(OpenAPIValidationError):
    pass



import sys
from os import path
from pathlib import Path
from typing import Any
from typing import Hashable
from typing import Mapping
from typing import Tuple

from jsonschema_path.handlers import all_urls_handler
from jsonschema_path.handlers import file_handler


def read_from_stdin(filename: str) -> Tuple[Mapping[Hashable, Any], str]:
    return file_handler(sys.stdin), ""  # type: ignore


def read_from_filename(filename: str) -> Tuple[Mapping[Hashable, Any], str]:
    if not path.isfile(filename):
        raise OSError(f"No such file: {filename}")

    filename = path.abspath(filename)
    uri = Path(filename).as_uri()
    return all_urls_handler(uri), uri
